# External Imports
from pydantic import BaseModel, Field, field_validator


class SimulationParameters(BaseModel):
    nx: int = Field(..., gt=0)
    ny: int = Field(..., gt=0)
    nz: int = Field(..., gt=0)
    dx: float = Field(..., gt=0)
    dy: float = Field(..., gt=0)
    dz: float = Field(..., gt=0)
    wind_speed: float = Field(..., ge=0)
    wind_direction: float = Field(..., ge=0, le=360)
    sim_time: int = Field(..., gt=0)
    auto_kill: int = Field(0, ge=0, le=1)
    num_cpus: int = Field(4, gt=0)
    fuel_flag: int = Field(...)
    ignition_flag: int = Field(...)
    output_time: int = Field(..., gt=0)
    topo_flag: int = Field(...)
    fuel_density: float = Field(None, ge=0)
    fuel_moisture: float = Field(None, ge=0)
    fuel_height: float = Field(None, ge=0)

    @field_validator("fuel_flag")
    def validate_fuel_flag(cls, v):
        assert v in (1, 3, 4, 5), "fuel_flag must be 1, 3, 4, or 5"
        return v

    @field_validator("topo_flag")
    def validate_topo_flag(cls, v):
        assert v in (0, 5), "topo_flag must be 0 or 5"
        return v

    @field_validator("ignition_flag")
    def validate_ignition_flag(cls, v):
        assert v in (1,
                     6), "ignition_flag must be 1 or 6. Future versions may support more options."
        return v

    @field_validator("fuel_density")
    def validate_fuel_density(cls, v, values):
        if values["fuel_flag"] == 1:
            assert v is not None, "fuel_density must be specified for fuel_flag=1"
        return v

    @field_validator("fuel_moisture")
    def validate_fuel_moisture(cls, v, values):
        if values["fuel_flag"] == 1:
            assert v is not None, "fuel_moisture must be specified for fuel_flag=1"
        return v

    @field_validator("fuel_height")
    def validate_fuel_height(cls, v, values):
        if values["fuel_flag"] == 1:
            assert v is not None, "fuel_height must be specified for fuel_flag=1"
        return v

    @classmethod
    def from_dict(cls, d):
        return cls(**d)

    def to_dict(self):
        return self.model_dump()

    @classmethod
    def from_fastfuels(cls, zarray, wind_speed, wind_direction, sim_time, auto_kill, num_cpus, fuel_flag, ignition_flag, output_time, topo_flag, fuel_density, fuel_moisture, fuel_height):
        return cls(nx=zarray.attrs["dx"],
                   ny=zarray.attrs["dy"],
                   nz=zarray.attrs["dz"],
                   dx=zarray.attrs["dx"],
                   dy=zarray.attrs["dy"],
                   dz=zarray.attrs["dz"],
                   wind_speed=wind_speed,
                   wind_direction=wind_direction,
                   sim_time=sim_time,
                   auto_kill=auto_kill,
                   num_cpus=num_cpus,
                   fuel_flag=fuel_flag,
                   ignition_flag=ignition_flag,
                   output_time=output_time,
                   topo_flag=topo_flag,
                   fuel_density=fuel_density,
                   fuel_moisture=fuel_moisture,
                   fuel_height=fuel_height)



"""
Module for converting QUIC-Fire output files to duck array data formats.
"""
from __future__ import annotations

# Core imports
import re
from pathlib import Path

# Internal imports
from quicfire_tools.parameters import SimulationParameters

# External imports
import zarr
import numpy as np
from numpy import ndarray

FUELS_OUTPUTS = {
    "fire-energy_to_atmos": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Energy released to the atmosphere that generates "
                       "buoyant plumes",
        "units": "kW",
    },
    "fire-reaction_rate": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Rate of reaction of fuel",
        "units": "kg/m^3/s",
    },
    "fuels-dens": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Fuel density",
        "units": "kg/m^3",
    },
    "fuels-moist": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Fuel moisture content",
        "units": "g water / g air",
    },
    "groundfuelheight": {
        "file_format": "gridded",
        "number_dimensions": 2,
        "grid": "fire",
        "delimiter": None,
        "extension": ".bin",
        "description": "2D array with initial fuel height in the ground layer",
        "units": "m",
    },
    "mburnt_integ": {
        "file_format": "gridded",
        "number_dimensions": 2,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "2D file containing the percentage of mass burnt for each (i,j) "
                       "location on the fire grid (vertically integrated)",
        "units": "%",
    },
}
THERMAL_RADIATION_OUTPUTS = {
    "thermaldose": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "The output shows the thermal flux to skin of a person "
                       "collocated with fuel (for health effects).",
        "units": "(kW/m^2)^(4/3)s",
    },
    "thermalradiation": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "The output shows the thermal flux to skin of a person "
                       "collocated with fuel (for health effects).",
        "units": "kW/m^2",
    },
}
WIND_OUTPUTS = {
    "windu": {
        "file_format": "gridded",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "",
        "extension": ".bin",
        "description": "Fire u-components, cell centered ",
        "units": "m/s",
    },
    "windv": {
        "file_format": "gridded",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "",
        "extension": ".bin",
        "description": "Fire v-components, cell centered ",
        "units": "m/s",
    },
    "windw": {
        "file_format": "gridded",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "",
        "extension": ".bin",
        "description": "Fire w-components, cell centered ",
        "units": "m/s",
    },
}
EMISSIONS_OUTPUTS = {
    "co-emissions": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Mass of CO emitted between two emission file output"
                       "times in grams",
        "units": "g",
    },
    "pm-emissions": {
        "file_format": "compressed",
        "number_dimensions": 3,
        "grid": "fire",
        "delimiter": "-",
        "extension": ".bin",
        "description": "Mass of PM2.5 emitted between two emission file output"
                       "times in grams",
        "units": "g",
    },
}
OUTPUTS_MAP = {
    **FUELS_OUTPUTS,
    **THERMAL_RADIATION_OUTPUTS,
    **WIND_OUTPUTS,
    **EMISSIONS_OUTPUTS,
}


class OutputFile:
    def __init__(self, name: str, file_format: str, shape: tuple,
                 grid: str, delimiter: str, extension: str,
                 description: str, units: str, index_map=None):
        self.name = name
        self.file_format = file_format
        self.shape = shape
        self.grid = grid
        self.delimiter = delimiter
        self.extension = extension
        self.description = description
        self.units = units
        self.times = []  # List of times corresponding to the timesteps
        self.filepaths = []  # List of file paths for each timestep
        function_mappings = {
            'gridded': _process_gridded_bin,
            'compressed': _process_compressed_bin
        }
        self._output_function = function_mappings.get(self.file_format)
        if self._output_function is None:
            raise ValueError(f"Unknown output format: {self.file_format}")
        self._compressed_index_map = index_map if (self.file_format ==
                                                   'compressed') else None

    def add_timestep(self, time: float, filepath: str) -> None:
        """Add a new timestep to the OutputFile."""
        self.times.append(time)
        self.filepaths.append(filepath)

    def to_numpy(self, timestep: int | list[int] | None = None) -> np.ndarray:
        self._validate_timestep(timestep)
        selected_files = self._select_files_based_on_timestep(timestep)
        return self._get_multiple_timesteps(selected_files)

    def _validate_timestep(self, timestep: int | list[int] | None) -> None:
        """Validate the timestep input."""
        if timestep is None:
            return
        if isinstance(timestep, int):
            if timestep not in range(len(self.times)):
                raise ValueError(f"Invalid timestep: {timestep}")
        elif all(isinstance(ts, int) for ts in timestep):
            if any(ts not in range(len(self.times)) for ts in timestep):
                raise ValueError(f"Invalid timestep: {timestep}")
        else:
            raise TypeError(f"Invalid timestep type: {type(timestep)}")

    def _select_files_based_on_timestep(self, timestep: int | list[int] | None) -> list[Path]:
        """Return files selected based on timestep."""
        if timestep is None:
            return self.filepaths
        if isinstance(timestep, int):
            return [self.filepaths[timestep]]
        return [self.filepaths[ts] for ts in timestep]

    def _get_single_timestep(self, output_file: Path) -> np.ndarray:
        """Return a numpy array for the given output file."""
        return self._output_function(output_file,
                                     self.shape,
                                     self._compressed_index_map)

    def _get_multiple_timesteps(self, output_files: list[Path]) -> np.ndarray:
        """Return a numpy array for the given output files."""
        arrays = [self._get_single_timestep(of) for of in output_files]
        return arrays[0] if len(arrays) == 1 else np.stack(arrays)


class SimulationOutputs:
    """
    A class responsible for managing and processing simulation outputs,
    including validation, extraction, and organization of data from output
    files. This class facilitates the retrieval of data in various formats
    and can return a numpy array for a specific output, or write the data to a
    zarr file.
    """

    def __init__(self, output_directory: Path | str,
                 params: SimulationParameters) -> None:
        # Convert to Path and resolve
        output_directory = Path(output_directory).resolve()

        # Validate outputs directory
        self._validate_output_dir(output_directory)

        # Assign attributes
        self.output_directory = output_directory
        self.params = params

        # Get indexing information from the fire grid
        self._fire_indexes = self._process_fire_indexes()

        # Build a list of present output files and their times
        self.outputs = {}
        self._build_output_files_map()

    @staticmethod
    def _validate_output_dir(outputs_directory):
        """
        Validate the outputs directory by checking for the existence of the
        Output directory and the required files.
        """
        # Check if the outputs directory exists and is a directory
        if not outputs_directory.exists():
            raise FileNotFoundError(
                f"The directory {outputs_directory} does not exist."
            )
        elif not outputs_directory.is_dir():
            raise NotADirectoryError(
                f"The path {outputs_directory} is not a directory."
            )

        # Check for required files
        required_files = ["fire_indexes.bin"]
        for f in required_files:
            if not (outputs_directory / f).exists():
                raise FileNotFoundError(
                    f"Required file {f} not found in outputs directory"
                )

    def _build_output_files_map(self):
        """
        Build a key-value map of output files, where the key is the name of
        the output file and the value is an instantiated OutputFile object.
        """
        for key, attributes in OUTPUTS_MAP.items():
            output_files_list = self._get_list_output_paths(key)

            # Check if any output files of the given type exist in the directory
            if output_files_list:
                shape = self._get_output_shape(key, attributes)
                self.outputs[key] = OutputFile(
                    name=key,
                    file_format=attributes["file_format"],
                    shape=shape,
                    grid=attributes["grid"],
                    delimiter=attributes["delimiter"],
                    extension=attributes["extension"],
                    description=attributes["description"],
                    units=attributes["units"],
                    index_map=self._fire_indexes
                )

                # Populate the output file with timesteps and filepaths
                for filepath in output_files_list:
                    time = self._get_output_file_time(key, filepath)
                    self.outputs[key].add_timestep(time, filepath)

    def _get_list_output_paths(self, name) -> list[Path]:
        """
        Get a sorted list of output files in the Output/ directory for the
        given output name.
        """
        paths = list(self.output_directory.glob(f"{name}*"))
        paths.sort()
        return paths

    def _get_output_shape(self, name, attrs) -> tuple:
        if attrs["number_dimensions"] == 2:
            return self.params.ny, self.params.nx
        # elif attrs["number_dimensions"] == 3 and name == "fire-energy_to_atmos":
        #     _process_vertical_grid(self.output_directory / "grid.bin", self.params.nx, self.params.ny)
        elif attrs["number_dimensions"] == 3 and attrs["grid"] == "fire":
            return self.params.ny, self.params.nx, self.params.nz + 1
        else:
            return self.params.ny, self.params.nx, self.params.nz

    @staticmethod
    def _get_output_file_time(output, fpath) -> int:
        """Get the time of the output file."""
        delimiter = OUTPUTS_MAP[output]['delimiter']
        extension = OUTPUTS_MAP[output]['extension']

        try:
            if delimiter == '':
                # Split on first numeric component
                name_parts = re.split(r'(\d+)', fpath.name)
                file_time = name_parts[1]
            else:
                file_time = fpath.name.split(delimiter)[-1].split(extension)[0]

            return int(file_time)

        except ValueError:  # No numeric component
            return 0

    def _process_fire_indexes(self) -> ndarray:
        """
        Process the `fire_indexes.bin` file to extract the necessary
        information to rebuild the 3D fields of fire and fuel related variables.

        - Data 1 (INT): Header
        - Data 2 (INT): Number of cells with fuel (firegrid%num_fuel_cells = nfuel below)
        - Data 3 (INT): Max index of the cells with fuel in the vertical direction
        - Data 4-(nfuel+3) (INT): Unique cell identifiers (firegrid%num_fuel_cells)
        - Data (nfuel+4)-(2*nfuel+4) (INT): i,j,k cell indexes
        - Data (INT): Header

        Function returns a 2D ndarray of shape (num_cells, 3) representing the
        i, j, k indexes of each cell.
        """
        with open(self.output_directory / "fire_indexes.bin", "rb") as fid:
            # Read in the number of cells and skip the header
            num_cells = np.fromfile(fid, dtype=np.int32, offset=4, count=1)[0]

            # Skip the max index of the cells with fuel in the vertical
            # direction and unique cell identifiers
            np.fromfile(fid, dtype=np.int32, count=7 + num_cells)

            # Initialize an array to hold the indices
            ijk = np.zeros((num_cells, 3), dtype=np.int32)

            # Loop over each dimension's indices (Like a sparse meshgrid)
            for i in range(0, 3):
                ijk[:, i] = np.fromfile(fid, dtype=np.int32, count=num_cells)

            # Convert to indices (numbered from 0)
            ijk -= 1

            return ijk

    def list_available_outputs(self):
        """Return a list of keys representing available outputs."""
        return list(self.outputs.keys())

    def get_output(self, key):
        """Return a list of times for a given output key."""
        return self.outputs[key]

    def to_numpy(self, key: str | OutputFile,
                 timestep: None | int | list[int] = None) -> np.ndarray:
        """Return a numpy array for the given output and timestep(s)."""
        output = self._validate_output(key)
        return output.to_numpy(timestep)

    def _validate_output(self, output: str | OutputFile) -> OutputFile:
        """Validate output."""
        if isinstance(output, str):
            try:
                output = self.outputs[output]
            except KeyError:
                raise ValueError(f"{output} is not a valid output. Valid "
                                 f"outputs are: {self.list_available_outputs()}")
        elif isinstance(output, OutputFile):
            if output not in self.outputs.values():
                raise ValueError(f"{output} is not a valid output. Valid "
                                 f"outputs are: {self.list_available_outputs()}")
        else:
            raise TypeError(f"output must be a string or OutputFile object. "
                            f"Got {type(output)}")
        return output

    def to_zarr(self, fpath: Path):
        """Write the data to a zarr file."""
        # Create the zarr file
        zarr_file = zarr.open(str(fpath), mode="w")

        # Write each output to the zarr file
        for output_name, output in self.outputs.items():
            # Create a zarr dataset for the output
            shape = (len(output.times), *output.shape)
            chunks = [1 if i == 0 else shape[i] for i in range(len(shape))]
            zarr_file.create_dataset(
                output_name,
                shape=shape,
                chunks=chunks,
                dtype=float)

            # Write each timestep to the output's zarr dataset
            for time_step in range(len(output.times)):
                data = self.to_numpy(output_name, time_step)
                zarr_file[output_name][time_step, ...] = data

        return zarr_file


def _process_compressed_bin(filename, dim_yxz, *args) -> ndarray:
    """
    Converts the contents of a sparse .bin file to a dense NumPy array.

    This function reads a .bin file containing sparse data and converts it to
    a dense 3D NumPy array. The mapping from sparse to dense is guided by an
    index map provided as an argument. The index map must match the sparse data
    in the .bin file, defining the locations of nonzero values within the dense
    array.

    Parameters
    ----------
    filename : str | Path
        The name or path of the sparse .bin file to be processed.
    dim_yxz : tuple of int
        The number of y, x, z cells to store in the 3D array, given as a tuple
        of three integers.
    *args : tuple
        Additional arguments, specifically expecting the index map for mapping
        sparse indices to dense. The index map should be a 2D array with shape
        (N, 3), where N is the number of nonzero elements in the sparse data.

    Returns
    -------
    ndarray
        A dense 3D NumPy array with shape defined by `dim_yxz`, and data type
        np.float32. The array represents the dense form of the sparse data
        from the .bin file.

    Raises
    ------
    IndexError
        If the index map (fire_indexes) is not provided in the arguments.
    ValueError
        If the index map (fire_indexes) is None.
   """
    # TODO: Rename fire_indexes to something more generic
    try:
        fire_indexes = args[0]
    except IndexError:
        raise ValueError("fire_indexes must be provided when processing "
                         "compressed .bin files.")
    if fire_indexes is None:
        raise ValueError("fire_indexes must be provided when processing "
                         "compressed .bin files.")

    # Initialize the 3D array
    full_3d_array = np.zeros(dim_yxz, dtype=np.float32)

    with open(filename, "rb") as fid:
        # Read header
        np.fromfile(fid, dtype=np.int32, count=1)

        # Read in the sparse values
        sparse_values = np.fromfile(
            fid, dtype=np.float32, count=fire_indexes.shape[0]
        )

        # Map indices of the sparse data to the indices of the dense array
        indices = (
            fire_indexes[:, 1],
            fire_indexes[:, 0],
            fire_indexes[:, 2],
        )

        # Update the full array with the sparse indices
        full_3d_array[indices] = sparse_values

    return full_3d_array


def _process_gridded_bin(filename, dims, *args) -> ndarray:
    """
    Converts the data stored in a gridded .bin file to an np array. The
    function handles both 2D and 3D data based on the dimensions provided.

    Parameters
    ----------
    filename : str | Path
        File path to the .bin file that contains the gridded data.
    dims : tuple
        Dimensions of the data. If a 2-element tuple (ny, nx) is provided,
        the function processes 2D data. If a 3-element tuple (ny, nx, nz)
        is provided, the function processes 3D data.
    *args : list
        Additional arguments that might be passed to the underlying slice
        processing function (_process_gridded_bin_slice).

    Returns
    -------
    ndarray
        A NumPy array containing the data extracted from the .bin file. The
        array's shape corresponds to the provided dimensions, and its data type
        is np.float32.
    """
    # Initialize the array
    array = np.zeros(dims, dtype=np.float32)

    with open(filename, "rb") as f:
        # Read header
        _ = np.fromfile(f, dtype=np.float32, count=1)

        # Read in gridded data
        if len(dims) == 2:  # 2D data
            array[:, :] = _process_gridded_bin_slice(f, *dims)
        else:  # 3D data
            for k in range(dims[2]):
                array[:, :, k] = _process_gridded_bin_slice(f, *dims[:2])
    return array


def _process_gridded_bin_slice(fid, ny, nx) -> ndarray:
    """
    Process a 2D slice from a gridded .bin file.

    This method reads a 2D slice of size (ny, nx) from the provided file
    object, and returns it as a np.float32 array.
    """
    # Read in gridded data
    plane = np.fromfile(fid, dtype=np.float32, count=ny * nx)
    return np.reshape(plane, (ny, nx))



from quicfire_tools.inputs import SimulationInputs
from quicfire_tools.outputs import SimulationOutputs, OutputFile
from quicfire_tools.parameters import SimulationParameters


__all__ = ['SimulationInputs',
           'SimulationParameters',
           'SimulationOutputs',
           'OutputFile']



"""
QUIC-Fire Tools Simulation Input Module
"""
from __future__ import annotations

# Core Imports
import time
from pathlib import Path
from string import Template

# Internal Imports
from quicfire_tools.parameters import SimulationParameters


class SimulationInputs:
    """
    Input Module
    """

    def __init__(self, directory: Path | str):
        if isinstance(directory, str):
            path = Path(directory)
            directory = path.resolve()

        directory.mkdir(parents=True, exist_ok=True)
        self.directory = directory

    def setup_input_files(self, params):
        """
        Populates input file templates with user defined parameters

        Parameters
        ----------
        params: SimulationParameters
            Dictionary of user defined parameters.

        Returns
        -------
        None:
            Sets up simulation files in the simulation directory.

        """
        # Convert params to dict
        params = params.to_dict()

        # Get current unix time
        params["timenow"] = int(time.time())

        # Write fuels data
        (params["fuel_density"], params["fuel_moisture"],
         params["fuel_height_flag"]) = self._write_fuel(params)

        # Write ignition data
        params["ignition_locations"] = self._write_ignition_locations(params)

        # Write input template files
        try:
            version = params["version"]
        except KeyError:
            version = "latest"
        template_files_path = Path(
            __file__).parent / "input-templates" / version
        template_files_list = template_files_path.glob("*")
        for fname in template_files_list:
            self._fill_form_with_dict(fname, params)

    @staticmethod
    def _write_fuel(params: dict) -> tuple[str, str, str]:
        """
        Writes fuel data to the QUIC_fire.inp input file. This function

        Parameters
        ----------
        params: dict
            Dictionary of user defined parameters.

        Returns
        -------
        None:
            Writes fuel data to the fuel_data.inp file.

        """
        # Get data from params
        fuel_flag = params["fuel_flag"]

        # Uniform fuel properties
        if fuel_flag == 1:
            fuel_density = "\n" + str(params["fuel_density"])
            fuel_moisture = "\n" + str(params["fuel_moisture"])
            fuel_height = (f"\n{fuel_flag}    ! fuel height flag: 1 = uniform; "
                           f"2 = provided thru QF_FuelMoisture.inp, 3 = Firetech"
                           f" files for quic grid, 4 = Firetech files for "
                           f"different grid (need interpolation)"
                           f"\n{str(params['fuel_height'])}")

        # Custom fuel .dat files (fuel flags 3 or 4)
        else:
            fuel_density, fuel_moisture, fuel_height = "", "", ""

        return fuel_density, fuel_moisture, fuel_height

    def _write_ignition_locations(self, params: dict) -> str:
        """
        Writes ignition locations file to simulation directory if ignition flag
        is 6. Otherwise, it writes the ignition locations to the QUIC_fire.inp
        input file.

        Parameters
        ----------
        params: dict
            Dictionary of user defined parameters.

        Returns
        -------
        ignition_locations: str
            Data string to write to the QUIC_fire.inp input file.

        """
        if params["ignition_flag"] == 6 or params["ignition_flag"] == 7:
            return ""

        else:  # Ignition flag 1
            return self._write_line_fire_ignition(params)

    @staticmethod
    def _write_line_fire_ignition(params: dict) -> str:
        """
        Writes line fire ignition locations to the QUIC_fire.inp input file.

        Line fire ignition takes the following parameters:
        fire_source_x0: south-west corner in the x-direction
        fire_source_y0: south-west corner in the y-direction
        fire_source_xlen: length in the x-direction
        fire_source_ylen: length in the y-direction
        num_ignition_cells: number of ignition cells constant at 100

        Parameters
        ----------
        params: dict
            Dictionary of user defined parameters.

        Returns
        -------
        ignition_locations: str
            Data string to write to the QUIC_fire.inp input file.
        """
        # Get data from params
        wind_direction = params["wind_direction"]
        nx, ny = params["nx"], params["ny"]
        dx, dy = params["dx"], params["dy"]

        # Hold number ignitions constant at 100 per documentation
        # num_ignition_cells = 100

        # Compute cardinal direction from wind direction
        dirs = ["N", "E", "S", "W"]
        ix = round(wind_direction / (360. / len(dirs)))
        cardinal_direction = dirs[ix % len(dirs)]

        # Ignition strip on top border
        if cardinal_direction == "N":
            fire_source_x0 = int(0.1 * nx * dx)
            fire_source_y0 = int(0.9 * ny * dy) - 1
            fire_source_xlen = int(0.8 * nx * dx)
            fire_source_ylen = 1

        # Ignition strip on left border
        elif cardinal_direction == "W":
            fire_source_x0 = int(0.1 * nx * dx) - 1
            fire_source_y0 = int(0.1 * ny * dy)
            fire_source_xlen = 1
            fire_source_ylen = int(0.8 * ny * dy)

        # Ignition strip on bottom border
        elif cardinal_direction == "S":
            fire_source_x0 = int(0.1 * nx * dx)
            fire_source_y0 = int(0.1 * ny * dy) - 1
            fire_source_xlen = int(0.8 * nx * dx)
            fire_source_ylen = 1

        # Ignition strip on right border
        else:
            fire_source_x0 = int(0.9 * nx * dx) - 1
            fire_source_y0 = int(0.1 * ny * dy)
            fire_source_xlen = 1
            fire_source_ylen = int(0.8 * ny * dy)

        ignition_list = ["", fire_source_x0, fire_source_y0, fire_source_xlen,
                         fire_source_ylen]
        return "\n".join((str(x) for x in ignition_list))

    def _fill_form_with_dict(self, template_file_path: Path, params: dict):
        """
        Fills a form with a dictionary of values.

        Parameters
        ----------
        template_file_path: Path
            Path to the template file
        params: dict
            Dictionary of user defined parameters.

        Returns
        -------
        None:
            Writes parameter data to the template file
        """
        output_file_path = self.directory / template_file_path.name
        with open(template_file_path, "r") as ftemp:
            with open(output_file_path, "w") as fout:
                src = Template(ftemp.read())
                result = src.substitute(params)
                fout.write(result)

        """
        Validates the params dictionary.

        Parameters
        ----------
        params: dict
            Dictionary of user defined parameters.

        Returns
        -------
        None:
            Raises ValueError if any parameters are invalid.
        """
        # Check for required parameters
        required_params = ["nx", "ny", "nz", "dx", "dy", "dz", "wind_speed",
                           "wind_direction", "sim_time", "num_cpus",
                           "fuel_flag", "ignition_flag", "output_time",
                           "topo_flag"]
        for param in required_params:
            if param not in params:
                raise KeyError(f"Parameter {param} is required.")

        # nx, ny, and nz must be integers greater than 0
        for param in ("nx", "ny", "nz"):
            if not isinstance(params[param], int):
                raise TypeError(f"Parameter {param} must be an integer.")
            if params[param] <= 0:
                raise ValueError(f"Parameter {param} must be greater than 0.")

        # dx, dy, and dz must be numbers greater than 0
        for param in ("dx", "dy", "dz"):
            if not isinstance(params[param], (int, float)):
                raise TypeError(f"Parameter {param} must be a number.")
            if params[param] <= 0:
                raise ValueError(f"Parameter {param} must be greater than 0.")

        # wind_speed must be a number greater than or equal to 0
        if not isinstance(params["wind_speed"], (int, float)):
            raise TypeError("Parameter wind_speed must be a number.")
        if params["wind_speed"] < 0:
            raise ValueError("Parameter wind_speed must be greater than or "
                             "equal to 0.")

        # wind_direction must be a number between 0 and 360
        if not isinstance(params["wind_direction"], (int, float)):
            raise TypeError("Parameter wind_direction must be a number.")
        if params["wind_direction"] < 0 or params["wind_direction"] > 360:
            raise ValueError("Parameter wind_direction must be between 0 and "
                             "360.")

        # sim_time must be an integer greater than 0
        if not isinstance(params["sim_time"], int):
            raise TypeError("Parameter sim_time must be an integer.")
        if params["sim_time"] <= 0:
            raise ValueError("Parameter sim_time must be greater than 0.")

        # auto_kill must be an integer equal to 0 or 1
        if not isinstance(params["auto_kill"], int):
            raise TypeError("Parameter auto_kill must be an integer.")
        if params["auto_kill"] not in (0, 1):
            raise ValueError("Parameter auto_kill must be 0 or 1.")

        # num_cpus must be an integer greater than 0
        if not isinstance(params["num_cpus"], int):
            raise TypeError("Parameter num_cpus must be an integer.")
        if params["num_cpus"] <= 0:
            raise ValueError("Parameter num_cpus must be greater than 0.")

        # output_time must be an integer greater than 0
        if not isinstance(params["output_time"], int):
            raise TypeError("Parameter output_time must be an integer.")
        if params["output_time"] <= 0:
            raise ValueError("Parameter output_time must be greater than 0.")

        # topo_flag must be an integer equal to 0 or 5
        if not isinstance(params["topo_flag"], int):
            raise TypeError("Parameter topo_flag must be an integer.")
        if params["topo_flag"] not in (0, 5):
            raise ValueError("Parameter topo_flag must be 0 or 5.")

        # Fuel flag must be an integer
        if not isinstance(params["fuel_flag"], int):
            raise TypeError("Parameter fuel_flag must be an integer.")

        # Fuel flags 1, 3, 4 are currently supported
        if params["fuel_flag"] not in (1, 3, 4, 5):
            raise ValueError("Parameter fuel_flag must be 1, 3, 4, or 5. Future"
                             "versions of this package will support more.")

        # If fuel_flag is 1, then the user must provide fuel_density,
        # fuel_moisture, and fuel_height parameters
        if params["fuel_flag"] == 1:
            if "fuel_density" not in params:
                raise KeyError("Parameter fuel_density is required when "
                               "fuel_flag is 1.")
            if "fuel_moisture" not in params:
                raise KeyError("Parameter fuel_moisture is required when "
                               "fuel_flag is 1.")
            if "fuel_height" not in params:
                raise KeyError("Parameter fuel_height is required when "
                               "fuel_flag is 1.")

        # fuel_density, fuel_moisture, and fuel_height must be numbers greater
        # or equal to 0
        for param in ("fuel_density", "fuel_moisture", "fuel_height"):
            if param in params:
                if not isinstance(params[param], (int, float)):
                    raise TypeError(f"Parameter {param} must be a number.")
                if params[param] < 0:
                    raise ValueError(f"Parameter {param} must be greater than"
                                     " or equal to 0.")

        # Ignition flag must be an integer
        if not isinstance(params["ignition_flag"], int):
            raise TypeError("Parameter ignition_flag must be an integer.")

        return params.copy()


if __name__ == '__main__':
    test_params = SimulationParameters(
        nx=100,
        ny=100,
        nz=1,
        dx=1.,
        dy=1.,
        dz=1.,
        wind_speed=4.,
        wind_direction=270,
        sim_time=60,
        auto_kill=1,
        num_cpus=1,
        fuel_flag=3,
        ignition_flag=1,
        output_time=10,
        topo_flag=0,
    )

    sim_test = SimulationInputs("../tests/data/test-simulation/")
    sim_test.setup_input_files(test_params)



import os

from setuptools import find_packages, setup


def read_file(fname):
    with open(fname, encoding="utf-8") as fd:
        return fd.read()


def get_version():
    with open(os.path.join("quicfire_tools", "_version.py")) as fd:
        # contents are __version__ = "<vstring>"
        return fd.read().split("=")[1].strip().strip('"')


NAME = "quicfire-tools"
DESCRIPTION = "Input and output management tools for the QUIC-Fire fire model"
LONG_DESCRIPTION = read_file("README.md")
VERSION = get_version()
LICENSE = "MIT"
URL = "https://github.com/silvxlabs/quicfire-tools"
PROJECT_URLS = {
    "Bug Tracker": f"{URL}/issues"
}

print(NAME, VERSION)

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    url=URL,
    project_urls=PROJECT_URLS,
    classifiers=["Intended Audience :: Developers",
                 "Intended Audience :: Science/Research",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "Programming Language :: Python :: 3.11",
                 "Topic :: Scientific/Engineering", ],
    package_dir={"": "."},
    packages=find_packages(exclude=["docs", "tests"]),
    package_data={
        'quicfire_tools': ['input-templates/'],
    },
    include_package_data=True,
    install_requires=[
        "numpy",
        "pydantic>=2",
        "zarr>=2",
    ],
    python_requires=">=3.8",
)
